#!/usr/bin/env perl

# Copyright (C) 2008 Ansgar Wiechers <ansgar.wiechers@planetcobalt.net>
#
# This module is free software. You may use, redistribute and/or modify it
# under the terms of the GNU General Public License Version 2.0 as found at
# <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>.

require 5.8.8;

use strict;
use warnings;

use lib '/usr/local/lib/perl';
use Getopt::Std;
use File::Basename qw( basename );
use Date::Manip;
use Sys::Syslog qw( :standard :macros );

use constant TRUE  => 1;
use constant FALSE => not TRUE;

# ==============================================================================

my $progname = basename($0);
my $options_string = 'cdf:hil';
my $date_format = '%Y-%m-%d %H:%M';
my $log_facility = LOG_MAIL;
my $debug = FALSE;
my $read_write = FALSE;
my $configname = 'sentdb.conf';
my @configfiles = ( 
		"/etc/$configname",
		"$ENV{'HOME'}/.$configname",
	);	

my %options = ();
my %config;
my $db;

# ==============================================================================

sub usage {
	print <<EOM;
Usage: $progname [-c|-h|-i|-l] [-d] [-f FILE]

       -c       cleanup database (remove entries older than maximum age)
       -d       run in debug mode
       -f FILE  read configuration from FILE
       -h       print this help
       -i       insert hashed parameters of mail into database
                (mail is read from STDIN)
       -l       list database contents

If the database file doesn't exist, it will be created automatically.

EOM
	exit 0;
}

# Uncomment the following lines if your version of Sys::Syslog doesn't provide
# the LOG_UPTO() function.
#sub LOG_UPTO {
#	my($pri) = @_;
#	eval q(((1<< (($pri)+1)) - 1));
#}

sub fail($) {
	syslog(LOG_ERR, $_[0]);
	closelog();
	die $_[0];
}

sub read_config {
	my @configfiles = @_;
	my %config;

	foreach my $filename (@configfiles) {
		next unless -f $filename;
		syslog(LOG_DEBUG, "Reading configuration from $filename.");
		open FILE, "<$filename" or fail("Cannot open $filename: $!");
		while (<FILE>) {
			next if m/^\s*#/ or m/^\s*$/; # skip comments and empty lines
			chomp;
			my ($key, $value) = split /\s*=\s*/;

			$key =~ s/(^\s*|\s*$)//g;
			$key =~ s/-/_/g;
			$key = lc($key);

			if (defined $value) {
				$value =~ s/(^\s*|\s*$)//g;
				$value =~ s/"(.*)"/$1/;
			}

			$config{$key} = $value;
		}	 
		close FILE;
	}

	return %config;
}

sub canonicalize_address_list($) {
	my $string = $_[0];
	my @list;

	$string =~ s/\s+/ /g;
	$string =~ s/(^ | $)//g;
	$string =~ s/^[a-z]+:\s+//i; # strip leading header name (from:, to:, ...)
	@list = split '\s*,\s*', $string;

	# strip aliases and stuff from e-mail addresses
	foreach my $item (@list) {
		$item =~ s/^(.*\s+)?<?(\S+@\S+).*$/$2/;
		$item =~ s/>$//;
	}

	return join(',', sort @list);
}

sub process_message {
	my (@content) = (@_);
	my ($from, $to, $date);

	for (my $i = 0; ($i <= $#content) and ($content[$i] !~ m/^\r?$/); $i++) {
		if ( $content[$i] =~ m/^from:/i ) {
			$from = $content[$i];
			$from .= $content[++$i] while $content[$i+1] =~ m/^\s+/;
		} elsif ( $content[$i] =~ m/^to:/i ) {
			$to = $content[$i];
			$to .= $content[++$i] while $content[$i+1] =~ m/^\s+/;
		} elsif ( $content[$i] =~ m/^date:\s+(.*)$/i ) {
			$date = $1;
		}
	}

	return undef unless defined $from and defined $to and defined $date;

	$from = canonicalize_address_list($from);
	$to = canonicalize_address_list($to);
	$date = ParseDate($date);

	return "$from;$to;$date";
}

# ==============================================================================

getopts($options_string, \%options);

usage(0) if $options{'h'};
$debug = TRUE if $options{'d'};
$read_write = TRUE if $options{'c'} or $options{'i'};
push(@configfiles, $options{'f'}) if $options{'f'};

if ($debug) {
	openlog($progname, 'ndelay,pid,perror', $log_facility);
	setlogmask(LOG_UPTO(LOG_DEBUG));
} else {
	openlog($progname, 'ndelay,pid', $log_facility);
	setlogmask(LOG_UPTO(LOG_INFO));
}

%config = read_config(@configfiles);

eval {
	if ($config{'backend'} eq 'bdb') {
		require FilterDB::BerkeleyDB;
		$db = new BerkeleyDB(
				'filename' => $config{'db_file'},
			);
	} elsif ($config{'backend'} eq 'sqlite') {
		require FilterDB::SQLite;
		$db = new SQLite(
				'filename' => $config{'db_file'},
			);
	} elsif ($config{'backend'} eq 'pgsql') {
		require FilterDB::PostgreSQL;
		$db = new PostgreSQL(
				'db_host'    => $config{'db_host'},
				'db_port'    => $config{'db_port'},
				'db_name'    => $config{'db_name'},
				'db_user'    => $config{'db_user'},
				'db_pass'    => $config{'db_pass'},
				'sslmode'    => $config{'sslmode'},
				'read_write' => $read_write,
			);
	} else {
		fail("Invalid database backend '$config{'backend'}'.");
	}
};
fail "$@" if $@;

if ($options{'c'}) {
	$db->cleanup();
} elsif ($options{'i'}) {
	my $digest = process_message(<>);
	fail('Incomplete message not added to database.') unless defined $digest;
	$db->insert($digest);
} elsif ($options{'l'}) {
	my %records = $db->list();
	my @keys = keys %records;
	if ($#keys >= 0) {
		foreach my $key (@keys) {
			print UnixDate($records{$key}, $date_format) . "\t$key\n";
		}
	} else {
		print STDERR "No entries in database.\n";
	}
}

closelog();

exit 0;
